/*-----------------------------------------------------------------------
		
	Inline Math library : Vector Matrix and Quaternion implementation
	
	* See imath.h for more information

	By Matt Barnett
		(see References in imath.h)
	
------------------------------------------------------------------------*/


#ifndef _IMATH_V_IMP_HEADER_4778097634v
#define _IMATH_V_IMP_HEADER_4778097634v

//-------------- ivector3 ------------
//
//  simple 3d vector with 3 components
//	(there is no 2d version, use this)
//------------------------------------

//enum for auto constructing some vector types
//if you want to contruct a random vector
//or  and random normalised vector etc.
enum ivector3_constructor_enum { IVECTOR3_RANDOM , IVECTOR3_RANDOM_NORM };

class ivector3
{
	public:
			
		//three floats are the data										
		union
		{
			float v[3];
			struct { float x, y, z; };
		};

		//-------------- operators ---
		//---------------------------- 
	    //return types according to Effective C++ 2e (see ref 8)
	
		inline const ivector3 operator-(const ivector3 &right) const 
		{
			return ivector3(x - right.x,y - right.y,z - right.z);
		}
		inline const ivector3 operator+(const ivector3 &right) const
		{
			return ivector3(x + right.x,y + right.y,z + right.z);
		}	
		
		inline const ivector3 operator*(const ivector3 &right) const 
		{
			return ivector3(x * right.x,y * right.y,z * right.z);
		}
		inline const ivector3 operator/(const ivector3 &right) const
		{
			return ivector3(x / right.x,y / right.y,z / right.z);
		}

		inline const ivector3 operator*(const float scalar) const 
		{
			return ivector3(x * scalar,y * scalar,z * scalar);
		}
		inline const ivector3 operator/(const float scalar) const 
		{
			return ivector3(x / scalar,y / scalar,z / scalar);
		}

		//void operators
		inline void operator-=(const ivector3 &right) 
		{
			subtract(right);
		}
		inline void operator+=(const ivector3 &right)
		{
			add(right);
		}

		//assignment
		inline ivector3 &operator=(const ivector3 &right)
		{
			if (&right != this) {				//check for self assign
				x = right.x;					//copy right to left					
				y = right.y;	
				z = right.z;	
			}
			return *this;						//enables x = y = z						
		}

		inline float operator[](int subscript) const
		{ 
			assert( subscript < 3 && subscript >= 0 );
			return v[subscript];
		}
		//---------------------------- 
		//---------------------------- 

		//------- construction etc ---
		//----------------------------
		
		//inlined ctor(initalises to 0,0,0)				
		inline ivector3()						   
		{
			set(0,0,0); 
		}					
		
		//inlined ctor(initalises to x,y,z)	
		inline ivector3(const float setx,const float sety,const float setz)
		{
			set(setx,sety,setz); 
		}	
		
		//ctor for a random normalised vector
		explicit inline ivector3(ivector3_constructor_enum inv_ce)
		{
			if (inv_ce == IVECTOR3_RANDOM_NORM)
			{
				randomize(1.0);
				normalise();
			}
			else if (inv_ce == IVECTOR3_RANDOM)
				randomize(1.0);
		}						

		//inlined dtor(unnessesary)
		inline ~ivector3()
		{}			
		//---------------------------- 
		//---------------------------- 

		//--------- void functions ---
		//----------------------------
		//sets vector components
		
		inline void set(const float setx,const float sety,const float setz)
		{
			x = setx;
			y = sety;
			z = setz;
		}	
		
		//sets z == 0
		inline void make2d()						  
		{ 
			z = 0;
		}															

		//dumps the vectors componets (debug aid)
		inline void cdump(const char *caption)  const 
		{
			cout << caption << x << " " << y << " " << z << endl;
		}	
		
		//draw O -> vector (debug aid)
		inline void gldraw() const			
		{
			glDisable(GL_LIGHTING);
			glBegin(GL_LINES);
			glVertex3f(0.0,0.0,0.0);
			glVertex3fv(v);
			glEnd();
		}														
		//plot the vector (debug aid)
		inline void glplot() const			
		{
			glVertex3fv(v);
		}	

		//multiplies this by scalar
		inline void multiply(const float scalar)			  
		{
			x *= scalar;
			y *= scalar;
			z *= scalar;
		}	
		
		//multiplies this by multiply
		inline void multiply(const ivector3 &multiply)   
		{ 
			x *= multiply.x;
			y *= multiply.y;
			z *= multiply.z; 
		}

		//divides this by scalar
    	inline void divide(const float scalar)			  
		{ 
			x /= scalar; 
			y /= scalar;
			z /= scalar;
		}
		
		//divides this by divide
    	inline void divide(const ivector3 &divide)			  
		{ 
			x /= divide.x;
			y /= divide.y;
			z /= divide.z; 
		}					
	
		//adds add to this		
		inline void add(const ivector3 &add)		
		{ 
			x += add.x;
			y += add.y;
			z += add.z; 
		}			
		
		//subtracts subtract from this
		inline void subtract(const ivector3 &subtract)	 	
		{
			x -= subtract.x;
			y -= subtract.y; 
			z -= subtract.z; 
		}

		//multiplies this by matrix multiply
		//inline void multiply(imatrix16 &multiply)  
		//{ 
		//
		//}	
		
		//normalise
		inline void normalise()			 
		{ 
			const float n = length();
			if (!n)
			{
				#ifdef _IVECTOR3_VERBOSE_NORMALISE_FAIL_
					printf("ivector3::MATH_ERROR ( normalise failed(), length = 0 )\n");
					cdump("vector ");
				#endif

				return;
			}

			x /= n;
			y /= n;
			z /= n; 
		}	
			
		//randomises each component
		inline void randomize(float range)
		{
			x = im_random_clamp(range);
			y = im_random_clamp(range);
			z = im_random_clamp(range);
		}
		
		//makes "this" the normal of the three vectors
		inline void makeTriangleNormal(const ivector3 &a,const ivector3 &b,const ivector3 &c)
		{
			ivector3 ab,ac;

			ab = b - a;
			ac = c - a;

			ivector3 cross = ab.cross(ac);
			cross.normalise();

			*this = cross;
			
		}
		//---------------------------- 
		//---------------------------- 
		
		//-------------- functions ---
		//----------------------------
		//crossproduct
		inline const ivector3 &cross(const ivector3 &in) const
		{
			ivector3 out;
			out.x = y*in.z - z*in.y; 
			out.y = z*in.x - x*in.z;
			out.z = x*in.y - y*in.x;
			return out;
		}
		
		//length/norm
		inline float length() const
		{
			float a = (x*x)+(y*y)+(z*z);

			if (a <= 0)
			{
				#ifdef _IVECTOR3_VERBOSE_LENGTH_FAIL_
					printf("ivector3::MATH_ERROR ( length failed, x*x+y*y+z*z <= 0 )\n");
					cdump("vector ");
					printf("x*x+y*y+z*z = %f\n",a);
				#endif

				return 0;
			}

			return im_sqrt(a); 
		}
		
		//length/norm
		inline float lengthsqr() const
		{
			float a = (x*x)+(y*y)+(z*z);

			if (a <= 0)
			{
				#ifdef _IVECTOR3_VERBOSE_LENGTH_FAIL_
					printf("ivector3::MATH_ERROR ( length failed, x*x+y*y+z*z <= 0 )\n");
					cdump("vector ");
					printf("x*x+y*y+z*z = %f\n",a);
				#endif

				return 0;
			}

			return a; 
		}
		
		//dotproduct
		inline float dot(const ivector3 &in) const
		{
			return x*in.x + y*in.y + z*in.z;
		}			
		
		//angle between this and in
		//
		//inline float angle(const ivector3 &in) const				
		//{	
		//	float tdotin,theta;
		//
		//	tdotin = dot(in);								//dot product
		//	
		//	float li;
		//	li = in.length();
		//
		//	theta = tdotin / (length() * li);				//lengths
		//	theta = (float)im_acos(theta);					//inverse im_cos
		//
		//	return theta;									//return theta in radians
		//}

		inline float cangle(const ivector3 &in) const
		{
			return ( x*in.x + y*in.y + z*in.z ) / ( (length() )*(in.length() ) );
		}

		//matrix based angle axis rotation (radians)
		//matrix isnt stored
		//usefull to rotate a point/vector about an axis
		inline void rotate(const float angle,const ivector3 &axis)
		{	
			//componets of rotation axis
			//float a = axis.x;		
			//float b = axis.y;
			//float c = axis.z;

			//get the matrix
			float f1cos,fasin,fbsin,fcsin;
			
			//fcos = im_cos(angle);		//im_cos of angle
			//fsin = im_sin(angle);		//im_sin of angle
			
			f1cos = (1 - im_cos(angle) );			//1 - im_cos of angle
			
			fasin = (axis.x * im_sin(angle) );			//a * im_sin of angle
			fbsin = (axis.y * im_sin(angle) );			//b * im_sin of angle
			fcsin = (axis.z * im_sin(angle) );			//c * im_sin of angle

			//m[0] = a*a*f1cos + fcos;
			//m[1] = a*b*f1cos - fcsin;
			//m[2] = a*c*f1cos + fbsin;

			//m[3] = a*b*f1cos + fcsin;
			//m[4] = b*b*f1cos + fcos;
			//m[5] = b*c*f1cos - fasin;

			//m[6] = a*c*f1cos - fbsin;
			//m[7] = b*c*f1cos + fasin;
			//m[8] = c*c*f1cos + fcos;

			ivector3 out;

			out.x = x * (axis.x * axis.x * f1cos + im_cos(angle) )  + y * (axis.x * axis.y * f1cos - fcsin) + z * (axis.x * axis.z * f1cos + fbsin);
			out.y = x * (axis.x * axis.y * f1cos + fcsin) + y * (axis.y * axis.y * f1cos + im_cos(angle) ) + z * (axis.y * axis.z * f1cos - fasin);
			out.z = x * (axis.x * axis.z * f1cos - fbsin) + y * (axis.y * axis.z * f1cos + fasin) + z * (axis.z * axis.z * f1cos + im_cos(angle) );

			x = out.x;
			y = out.y;
			z = out.z;

		}
		//---------------------------- 
		//---------------------------- 
};

//------ ivector3 constants -----
//-------------------------------

const ivector3 xaxis(1.0,0.0,0.0);
const ivector3 yaxis(0.0,1.0,0.0);
const ivector3 zaxis(0.0,0.0,1.0);
const ivector3 zero (0.0,0.0,0.0);

//a few colors
const ivector3 red  (1.0,0.0,0.0);  
const ivector3 green(0.0,1.0,0.0);
const ivector3 blue (0.0,0.0,1.0);
const ivector3 black(0.0,0.0,0.0);

const ivector3 cyan   (0.0,1.0,1.0);
const ivector3 magenta(1.0,0.0,1.0);
const ivector3 yellow (1.0,1.0,0.0);

//----------------------------
//----------------------------

//-------------- ibasis9 ------
//
// just for storing and drawing
// orthonominal(??) bases or axes
//
// there is no clever shit here making
// sure that the vectors remain perpendicular
// this is just for storage of 3 vectors.

class ibasis9
{
	public:	
		
		//three vectors are the data
		ivector3 x,y,z;

		ibasis9::ibasis9(const ivector3 &inx,const ivector3 &iny,const ivector3 &inz)
		{
			x = inx;
			y = iny;
			z = inz;
		}
		
		ibasis9::ibasis9()
		{
			x = xaxis;
			y = yaxis;
			z = zaxis;
		}

		//just normalises each vector to save some typing
		inline void normalise()
		{
			x.normalise();
			y.normalise();
			z.normalise();
		}

		//draws the basis (debug aid)
		inline void gldraw() const
		{
			glDisable(GL_LIGHTING);
			glBegin(GL_LINES);
			glColor3fv(cyan.v);   glVertex3fv(zero.v);glVertex3fv(x.v);
			glColor3fv(magenta.v);glVertex3fv(zero.v);glVertex3fv(y.v);
			glColor3fv(yellow.v); glVertex3fv(zero.v);glVertex3fv(z.v);
			glEnd();
		}
};

//---------------------------- 
//---------------------------- 

#endif
